/*
 * TXX -- A Virtual Tcode Machine
 * Copyright (C) 1997-2004 Nils M Holm. All rights reserved.
 * See the file LICENSE for conditions of use.
 *
 * SYNTAX: TXX PROGRAM [OPTIONS]
 */

#ifdef plan9
 #include <u.h>
 #include <libc.h>
 #define CREATE(p, m)	create(p, OWRITE, m)
 #define EXT_SYSTEM
 #define EXT_TTYCTL
#else
 #define CREATE(p, m)	creat(p, m)
#endif

#include <stdio.h>

#ifndef plan9
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#endif

#ifdef __TURBOC__
 #include <alloc.h>
 #include <io.h>
#endif

#ifndef __TURBOC__
 #ifndef plan9
  #include <unistd.h>
 #endif
#endif

#define MAIN
#include "txx.h"
#include "tcode.h"
#include "rtcode.h"

void	ifail(void);

#ifdef dos
 #define TEXTSIZE	(32000u * BPW)
 #define DATASIZE	TEXTSIZE
 #define STACKSIZE	1024
 #define NLABELS	4096
 #define NMARKS		4096
#else
 #define TEXTSIZE	(65536u * BPW)
 #define DATASIZE	TEXTSIZE
 #define STACKSIZE	32768u
 #define NLABELS	8192
 #define NMARKS		8192
#endif

#define LDBUFLEN	1024

struct mark {
	int	l, t, d;
};

#define MAXINFS		256

struct inf {
	char	*name;
	UINT	cmap;
	int	narg;
	XCELL	(*proc)();
};

/* shared variables are defined in 'txx.h' */
UCHAR		*Text;
CELL		*Wdata;
UCELL		*Uwdata;
UCELL		Etext, Edata;
UCELL		Datalen = DATASIZE,
		Textlen = TEXTSIZE;
int		Tags[NLABELS];
struct mark	Marks[NMARKS];
UINT		Mt;
UINT		Dptr, Tptr;
UINT		Op;
CELL		A1, A2;
UCELL		U1;
UCELL		Ip, Fp;
UCELL		Self;

UCHAR		Ldbuf[LDBUFLEN];
int		Lfd, Lp, Lz;

extern struct infproc	InfT3X[];
#ifdef EXT_SYSTEM
extern struct infproc	InfSystem[];
#endif
#ifdef EXT_TTYCTL
extern struct infproc	InfTTYCtl[];
#endif

struct inf	Infs[MAXINFS], *Lastinf = NULL;

struct infproc	*Interfaces[] = {
	InfT3X,
#ifdef EXT_SYSTEM
	InfSystem,
#endif
#ifdef EXT_TTYCTL
	InfTTYCtl,
#endif
	NULL };

#ifdef plan9
void exit(int n) {
	char	buf[10];

	if (n==0) exits(NULL);
	sprintf(buf, "%d", n);
	exits(buf);
}
#endif

void coredump(void) {
	int	fd;

	fd = CREATE("TXXCODE", 0600);
	if (fd < 0) return;
	write(fd, Text, Etext);
	close(fd);
	fd = CREATE("TXXDATA", 0600);
	if (fd < 0) return;
	write(fd, Data, Edata);
	close(fd);
	fprintf(stderr, "TXX core dumped\n");
}

void fail(char *m, char *n) {
	int	nodump = 0;

	if (Etext == 0) {
		Etext = Tptr;
		nodump = 1;
	}
	if (n)
		fprintf(stderr, "TXX core fault: %s: %s\n", m, n);
	else
		fprintf(stderr, "TXX core fault: %s\n", m);
	fprintf(stderr, "IP=%u, SP=%u, FP=%u, RR=%d, SELF=%u\n"
			"Etext=%u, Edata=%u\n",
		Ip, Sp<<XBPW, Fp<<XBPW, Acc, Self, Etext, Edata);
	if (Op) {
		fprintf(stderr, "OP=%s A1=%d (%u) A2=%d (%u)\n",
			Tcode_Opcodes[Op], A1, U1, A2, A2);
	}
	if (!nodump) coredump();
	exit(2);
}

void init(void) {
	int	i;

	Text = (UCHAR *) malloc(TEXTSIZE);
	Data = (UCHAR *) malloc(DATASIZE);
	if (!Text || !Data) fail("insufficient memory", 0);
	memset(Data, 0, DATASIZE);
	Wdata = (CELL *) Data;
	Stack = Wdata;
	Uwdata = (UCELL *) Data;
	Ustack = Uwdata;
	Sp = Fp = DATASIZE>>XBPW;
	Etext = 0;
	Edata = 0;
	for (i=0; i<NLABELS; i++) Tags[i] = -1;
	Mt = 0;
	Tptr = 0;
	Dptr = 0;
	A1 = 0;
	A2 = 0;
	Op = 0;
	Self = 0;
	libinit();
}

void addlabel(int op, int id) {
	UINT	a, i, o;
	UINT	here;

	if (id >= NLABELS) fail("label too big", 0);
	if (Tags[id] != -1) fail("duplicate label", 0);
	here = op == (ICLAB & IOPMASK)? Tptr: Dptr;
	Tags[id] = here;
	for (i=0; i<Mt; i++) if (Marks[i].l == id) break;
	if (i >= Mt) return;
	a = Marks[i].t;
	while (a) {
		o = a;
		a = (Text[a] << 24) | (Text[a+1] << 16) |
			(Text[a+2] << 8) | Text[a+3];
		Text[o] = here >> 24;
		Text[o+1] = here >> 16;
		Text[o+2] = here >> 8;
		Text[o+3] = here;
	}
	a = Marks[i].d;
	while (a) {
		o = a;
		a = (Data[a] << 24) | (Data[a+1] << 16) |
			(Data[a+2] << 8) | Data[a+3];
		Data[o] = here >> 24;
		Data[o+1] = here >> 16;
		Data[o+2] = here >> 8;
		Data[o+3] = here;
	}
}

int addmark(int id, int dref) {
	int	i, old;

	for (i=0; i<Mt; i++) {
		if (Marks[i].l == id) {
			if (dref) {
				old = Marks[i].d;
				Marks[i].d = Dptr;
			}
			else {
				old = Marks[i].t;
				Marks[i].t = Tptr;
			}
			return(old);
		}
	}
	if (Mt >= NMARKS) fail("too many marks", 0);
	Marks[Mt].l = id;
	if (dref) {
		Marks[Mt].d = Dptr;
		Marks[Mt].t = 0;
	}
	else {
		Marks[Mt].t = Tptr;
		Marks[Mt].d = 0;
	}
	Mt++;
	return(0);
}

int findlab(int id, int dref) {
	if (Tags[id] != -1) return(Tags[id]);
	return(addmark(id, dref));
}

UCHAR rdc(void) {
	if (Lp >= Lz) {
		Lz = read(Lfd, Ldbuf, LDBUFLEN);
		Lp = 0;
	}
	return(Ldbuf[Lp++]);
}

UCELL rdop(int k) {
	UCELL	n;

	n = rdc();
	if (k == 0) return n>0x7F? (-1 & ~0x7F) | n: n;
	n = (n << 8) | rdc();
	if (k == 1) return n>0x7FFF? (-1 & ~0x7FFF) | n: n;
	/* if (k != 3) fail("RDOP: bad argument length", NULL); */
	n = (n << 8) | rdc();
	n = (n << 8) | rdc();
#ifdef HOST16
	return n;
#else
	return n>0x7FFFFFFFL? (-1L & ~0x7FFFFFFFL) | n: n;
#endif
}

void dgrow(int n) {
	int	oldlen;

#ifdef HOST16
	if ((Dptr += n) > DATASIZE-STACKSIZE)
		fail("data segment overflow", 0);
#else
	if ((Dptr += n) <= Datalen-STACKSIZE) return;
	oldlen = Datalen;
	while (Datalen-STACKSIZE < Dptr) Datalen += DATASIZE;
	Data = (UCHAR *) realloc(Data, Datalen);
	if (!Data) fail("insufficient memory", 0);
	memset(&Data[oldlen], 0, Datalen-oldlen);
	Wdata = (CELL *) Data;
	Stack = Wdata;
	Uwdata = (UCELL *) Data;
	Ustack = Uwdata;
	Sp = Fp = Datalen>>XBPW;
#endif
}

void tgrow(int n) {
#ifdef HOST16
	if ((Tptr += n) > TEXTSIZE)
		fail("text segment overflow", 0);
#else
	if ((Tptr += n) <= Textlen) return;
	Textlen += TEXTSIZE;
	Text = (UCHAR *) realloc(Text, Textlen);
	if (!Text) fail("insufficient memory", 0);
#endif
}

void wrop(int arg, int k) {
	switch (k) {
	case 0:	Text[Tptr] = arg;
		tgrow(1);
		break;
	case 1:	Text[Tptr] = arg >> 8;
		Text[Tptr+1] = arg;
		tgrow(2);
		break;
	case 3:	Text[Tptr] = arg >> 24;
		Text[Tptr+1] = arg >> 16;
		Text[Tptr+2] = arg >> 8;
		Text[Tptr+3] = arg;
		tgrow(4);
		break;
	default:
		fail("WROP: bad argument length", NULL);
	}
}

void skip(int n) {
	int	i;

	for (i=0; i<n; i++) rdc();
}

void copy(int n, UCHAR *p) {
	int	i;

	for (i=0; i<n; i++) p[i] = rdc();
}

void addinf(int id, char *name, int len) {
	int	i, j;

	name[len] = 0;
	if (id >= MAXINFS)
		fail("interface ID out of range", name);
	if (Infs[id].name) {
		if (strcmp(Infs[id].name, name))
			fail("internal error: interface collision", name);
		return;
	}
	if ((Infs[id].name = (char *) malloc(len+1)) == NULL)
		fail("insufficient memory", 0);
	memcpy(Infs[id].name, name, len+1);
	for (j=0; Interfaces[j]; j++) {
		for (i=0; Interfaces[j][i].name; i++) {
			if (!strcmp(Interfaces[j][i].name, name)) {
				Infs[id].proc = Interfaces[j][i].proc;
				Infs[id].cmap = 0;
				Infs[id].narg = 0;
				Lastinf = &Infs[id];
				return;
			}
		}
	}
	fail("unsupported interface", name);
}

int resolve(void) {
	UCELL	op, na, fl, arg, arg2;
	UINT	b;
	char	buf[257];

	arg = arg2 = 0;	/* lint */
	op = rdc();
	if (op == '#') {
		while (Ldbuf[Lp++] != '\n')
			if (Lp >= LDBUFLEN-1)
				fail("#! header too long", 0);
	}
	else {
		Lp--;
	}
	if (Ldbuf[Lp+1] != (IINIT & IOPMASK))
		fail("bad input program (magic match failed)", 0);
	for (;;) {
		op = rdc();
		if (Lz < 1) break;
		if (op & IARFLAG) {
			fl = op;
			op = rdc();
			na = (fl & IARMASK) >> IAMOFF;
			if (na) {
				arg = rdop((fl & IL1MASK) >> IL1OFF);
				if (na > 1)
					arg2 = rdop((fl & IL2MASK) >> IL2OFF);
			}
		}
		else {
			fl = na = 0;
		}
		switch (op) {
		case ICLAB & IOPMASK:
		case IDLAB & IOPMASK:
			addlabel(op, arg);
			op = 0;
			break;
		case IVEC & IOPMASK:
			dgrow(arg? (arg<<XBPW): BPW);
			op = 0;
			break;
		case IDATA & IOPMASK:
			Wdata[Dptr>>XBPW] = arg;
			dgrow(BPW);
			op = 0;
			break;
		case ICREF & IOPMASK:
		case IDREF & IOPMASK:
			Wdata[Dptr>>XBPW] = findlab(arg, 1);
			dgrow(BPW);
			op = 0;
			break;
		case ISTR & IOPMASK:
			b = Dptr;
			dgrow(arg);
			copy(arg, &Data[b]);
			do {
				Data[Dptr++] = 0;
			} while (Dptr & (BPW-1));
			op = 0;
			break;
		case IINCG & IOPMASK:
		case ILDG & IOPMASK:
		case ILDGV & IOPMASK:
		case ISAVG & IOPMASK:
		case ILDLAB & IOPMASK:
		case ICALL & IOPMASK:
		case IBRF & IOPMASK:
		case IBRT & IOPMASK:
		case INBRF & IOPMASK:
		case INBRT & IOPMASK:
		case IJUMP & IOPMASK:
		case IUNEXT & IOPMASK:
		case IDNEXT & IOPMASK:
			fl = (fl & ~(IL1MASK<<IL1OFF)) | (3 << IL1OFF);
			Text[Tptr] = fl;
			Text[Tptr+1] = op;
			tgrow(2);
			arg = findlab(arg, 0);
			break;
		case IPUB & IOPMASK:
			skip(arg2);
			op = 0;
			break;
		case ILSYM & IOPMASK:
		case IGSYM & IOPMASK:
		case IISYM & IOPMASK:
			skip(arg2);
			op = 0;
			break;
		case IEXT & IOPMASK:
			copy(arg2, Text);
			Text[arg2] = 0;
			fail("unresolved external reference", (char *) Text);
			break;
		case IIREF & IOPMASK:
			copy(arg2, Text);
			Text[arg2] = 0;
			fail("unresolved interface reference", (char *) Text);
			break;
		case IHINT & IOPMASK:
			op = 0;
			break;
		case IINIT & IOPMASK:
			if (arg != 8) fail("unsupported tcode version", NULL);
			op = 0;
			break;
		case IBPW & IOPMASK:
			if (arg != BPW)
				fail("unsupported machine word size", NULL);
			break;
		case ILIB & IOPMASK:
			skip(arg);
			op = 0;
			break;
		case IIPROC & IOPMASK:
			copy(arg2, (UCHAR *) buf);
			addinf(arg, buf, arg2);
			op = 0;
			break;
		case ICMAP & IOPMASK:
			if (!Lastinf) fail("CMAP without IPROC", 0);
			Lastinf->narg = arg;
			Lastinf->cmap = arg2;
			op = 0;
			break;
		default:
			if (fl) {
				Text[Tptr] = fl;
				tgrow(1);
			}
			Text[Tptr] = op;
			tgrow(1);
			break;
		}
		if (op && na) {
			wrop(arg, (fl & IL1MASK) >> IL1OFF);
			if (na > 1) wrop(arg2, (fl & IL2MASK) >> IL2OFF);
		}
	}
	Etext = Tptr;
	Edata = Dptr;
	return 0;
}

void load(char *name) {
	if ((Lfd = open(name, 0)) < 0) fail("cannot open program", name);
	Lp = Lz = 0;
	resolve();
	close(Lfd);
}

void infcall(UINT id) {
	XCELL	(*p0)(void),
		(*p1)(XCELL),
		(*p2)(XCELL, XCELL),
		(*p3)(XCELL, XCELL, XCELL),
		(*p4)(XCELL, XCELL, XCELL, XCELL),
		(*p5)(XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p6)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p7)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p8)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL),
		(*p9)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL),
		(*p10)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL),
		(*p11)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL),
		(*p12)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p13)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p14)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL),
		(*p15)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL),
		(*p16)(XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL, XCELL, XCELL, XCELL, XCELL, XCELL,
			XCELL, XCELL);

	#define	P0
	#define	P1	P[0]
	#define	P2	P[1],P[0]
	#define	P3	P[2],P[1],P[0]
	#define	P4	P[3],P[2],P[1],P[0]
	#define	P5	P[4],P[3],P[2],P[1],P[0]
	#define	P6	P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P7	P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P8	P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P9	P[8],P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P10	P[9],P[8],P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P11	P[10],P[9],P[8],P[7],P[6],P[5],P[4],P[3],P[2], \
			P[1],P[0]
	#define	P12	P[11],P[10],P[9],P[8],P[7],P[6],P[5],P[4],P[3], \
			P[2],P[1],P[0]
	#define	P13	P[12],P[11],P[10],P[9],P[8],P[7],P[6],P[5],P[4], \
			P[3],P[2],P[1],P[0]
	#define	P14	P[13],P[12],P[11],P[10],P[9],P[8],P[7],P[6],P[5], \
			P[4],P[3],P[2],P[1],P[0]
	#define	P15	P[14],P[13],P[12],P[11],P[10],P[9],P[8],P[7],P[6], \
			P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P16	P[15],P[14],P[13],P[12],P[11],P[10],P[9],P[8],P[7], \
			P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	XCELL	P[16];
	int	i, n;
	UINT	b, m;

	if (id >= MAXINFS || Infs[id].name == NULL)
		fail("bad interface call", 0);
	n = Infs[id].narg;
	m = Infs[id].cmap;
	for (b=1,i=0; i<n; i++) {
		if (m & b) {
			P[i] = Ustack[Sp+n-i];
			/* The next statement relies upon
			 * sizeof(XCELL) == sizeof(char*).
			 * It may cause a warning.
			*/
			P[i] = (XCELL) &Data[P[i]];
		}
		else {
			P[i] = Stack[Sp+n-i];
		}
		b <<= 1;
	}
	switch(n) {
	case  0: p0 = Infs[id].proc; Acc = (*p0)(P0); break;
	case  1: p1 = Infs[id].proc; Acc = (*p1)(P1); break;
	case  2: p2 = Infs[id].proc; Acc = (*p2)(P2); break;
	case  3: p3 = Infs[id].proc; Acc = (*p3)(P3); break;
	case  4: p4 = Infs[id].proc; Acc = (*p4)(P4); break;
	case  5: p5 = Infs[id].proc; Acc = (*p5)(P5); break;
	case  6: p6 = Infs[id].proc; Acc = (*p6)(P6); break;
	case  7: p7 = Infs[id].proc; Acc = (*p7)(P7); break;
	case  8: p8 = Infs[id].proc; Acc = (*p8)(P8); break;
	case  9: p9 = Infs[id].proc; Acc = (*p9)(P9); break;
	case 10: p10 = Infs[id].proc; Acc = (*p10)(P10); break;
	case 11: p11 = Infs[id].proc; Acc = (*p11)(P11); break;
	case 12: p12 = Infs[id].proc; Acc = (*p12)(P12); break;
	case 13: p13 = Infs[id].proc; Acc = (*p13)(P13); break;
	case 14: p14 = Infs[id].proc; Acc = (*p14)(P14); break;
	case 15: p15 = Infs[id].proc; Acc = (*p15)(P15); break;
	case 16: p16 = Infs[id].proc; Acc = (*p16)(P16); break;
	default: fail("too many arguments in ICALL", Infs[id].name); break;
	}
}

void divz(void) { fail("divide by zero", 0); }

void ifail(void) { Op = 0; fail("illegal opcode", 0); }

void iadd(void) { Stack[Sp+1] += Stack[Sp]; Sp++; }
void iband(void) { Stack[Sp+1] &= Stack[Sp]; Sp++; }
void ibnot(void) { Stack[Sp] = ~Stack[Sp]; }
void ibor(void) { Stack[Sp+1] |= Stack[Sp]; Sp++; }
void ibpw(void) { ifail(); }
void ibrf(void) { if (!Stack[Sp++]) Ip = U1; }
void ibrt(void) { if (Stack[Sp++]) Ip = U1; }
void ibshl(void) { Stack[Sp+1] <<= Stack[Sp]; Sp++; }
void ibshr(void) { Ustack[Sp+1] >>= Ustack[Sp]; Sp++; }
void ibxor(void) { Stack[Sp+1] ^= Stack[Sp]; Sp++; }
void icall(void) { Stack[--Sp] = Ip; Ip = U1; }
void icalr(void) { UCELL t; t = Stack[Sp]; Stack[Sp] = Ip; Ip = t; }
void icalx(void) { fail("call to external procedure", 0); }
void iclab(void) { ifail(); }
void iclean(void) { Sp += A1; Stack[--Sp] = Acc; }
void icmap(void) { ifail(); }
void icref(void) { ifail(); }
void idata(void) { ifail(); }
void ideref(void) { Stack[Sp+1] += (Stack[Sp]<<XBPW); Sp++;
	Stack[Sp] = Wdata[Ustack[Sp]>>XBPW]; }
void idiv(void) { if (!Stack[Sp]) divz(); Stack[Sp+1] /= Stack[Sp]; Sp++; }
void idlab(void) { ifail(); }
void idnext(void) { if (Stack[Sp+1] <= Stack[Sp]) Ip = U1; Sp+=2; }
void idref(void) { ifail(); }
void idrefb(void) { Stack[Sp+1] += Stack[Sp]; Sp++; Stack[Sp] = Data[Ustack[Sp]]; }
void idup(void) { Stack[Sp-1] = Stack[Sp]; Sp--; }
void iend(void) { Fp = Stack[Sp++]; Ip = Stack[Sp++]; }
void iendm(void) { Self = Stack[Sp++]; Fp = Stack[Sp++]; Ip = Stack[Sp++]; }
void iequ(void) { Stack[Sp+1] = Stack[Sp+1] == Stack[Sp]? -1: 0; Sp++; }
void iext(void) { ifail(); }
void iglue(void) { }
void igrtr(void) { Stack[Sp+1] = Stack[Sp+1] > Stack[Sp]? -1: 0; Sp++; }
void igsym(void) { ifail(); }
void igteq(void) { Stack[Sp+1] = Stack[Sp+1] >= Stack[Sp]? -1: 0; Sp++; }
void ihalt(void) { exit(A1); }
void ihdr(void) { Stack[--Sp] = Fp; Fp = Sp; }
void ihint(void) { ifail(); }
void iicall(void) { infcall(U1); }
void iicalx(void) { fail("call to external interface", 0); }
void iincg(void) { Wdata[U1>>XBPW] += A2; }
void iinci(void) { Wdata[(Self>>XBPW)+U1] += A2; }
void iincl(void) { Stack[Fp-A1] += A2; }
void iinit(void) { ifail(); }
void iiproc(void) { ifail(); }
void iiref(void) { ifail(); }
void iisym(void) { ifail(); }
void ijump(void) { Ip = U1; }
void ildg(void) { Stack[--Sp] = Wdata[U1>>XBPW]; }
void ildgv(void) { Stack[--Sp] = U1; }
void ildi(void) { Stack[--Sp] = Wdata[(Self>>XBPW) + U1]; }
void ildiv(void) { Stack[--Sp] = Self + (U1<<XBPW); }
void ildl(void) { Stack[--Sp] = Stack[Fp-A1]; }
void ildlab(void) { Stack[--Sp] = U1; }
void ildlv(void) { Stack[--Sp] = (Fp-A1)<<XBPW; }
void iless(void) { Stack[Sp+1] = Stack[Sp+1] < Stack[Sp]? -1: 0; Sp++; }
void ilib(void) { ifail(); }
void iline(void) { }
void ilnot(void) { Stack[Sp] = Stack[Sp]? 0: -1; }
void ilsym(void) { ifail(); }
void ilteq(void) { Stack[Sp+1] = Stack[Sp+1] <= Stack[Sp]? -1: 0; Sp++; }
void imhdr(void) { Stack[--Sp] = Fp; Fp = Sp;
	Stack[--Sp] = Self; Self = Stack[Sp+3]; }
void imod(void) { if (!Stack[Sp]) divz(); Ustack[Sp+1] %= Ustack[Sp]; Sp++; }
void imul(void) { Stack[Sp+1] *= Stack[Sp]; Sp++; }
void inbrf(void) { if (!Stack[Sp]) Ip = U1; }
void inbrt(void) { if (Stack[Sp]) Ip = U1; }
void ineg(void) { Stack[Sp] = -Stack[Sp]; }
void inequ(void) { Stack[Sp+1] = Stack[Sp+1] != Stack[Sp]? -1: 0; Sp++; }
void inorm(void) { Stack[Sp+1] += (Stack[Sp]<<XBPW); Sp++; }
void inormb(void) { Stack[Sp+1] += Stack[Sp]; Sp++; }
void inum(void) { Stack[--Sp] = A1; }
void ipop(void) { Acc = Stack[Sp++]; }
void ipstr(void) { ifail(); }
void ipub(void) { ifail(); }
void isavg(void) { Wdata[U1>>XBPW] = Stack[Sp++]; }
void isavi(void) { Wdata[(Self>>XBPW)+U1] = Stack[Sp++]; }
void isavl(void) { Stack[Fp-A1] = Stack[Sp++]; }
void iself(void) { Stack[--Sp] = Self; }
void istack(void) { Sp -= A1; if ((Sp<<XBPW) < Edata) fail("stack overflow", 0); }
void istorb(void) { Data[Ustack[Sp+1]] = Ustack[Sp]; Sp+=2; }
void istore(void) { Wdata[Ustack[Sp+1]>>XBPW] = Stack[Sp]; Sp+=2; }
void istr(void) { ifail(); }
void isub(void) { Stack[Sp+1] -= Stack[Sp]; Sp++; }
void iswap(void) { UCELL t; t = Stack[Sp+1]; Stack[Sp+1] = Stack[Sp];
	Stack[Sp] = t; }
void iudiv(void) { if (!Stack[Sp]) divz(); Ustack[Sp+1] /= Ustack[Sp]; Sp++; }
void iugrtr(void) { Ustack[Sp+1] = Ustack[Sp+1] > Ustack[Sp]? -1: 0; Sp++; }
void iugteq(void) { Ustack[Sp+1] = Ustack[Sp+1] >= Ustack[Sp]? -1: 0; Sp++; }
void iuless(void) { Ustack[Sp+1] = Ustack[Sp+1] < Ustack[Sp]? -1: 0; Sp++; }
void iulteq(void) { Ustack[Sp+1] = Ustack[Sp+1] <= Ustack[Sp]? -1: 0; Sp++; }
void iumul(void) { Ustack[Sp+1] *= Ustack[Sp]; Sp++; }
void iunext(void) { if (Stack[Sp+1] >= Stack[Sp]) Ip = U1; Sp+=2; }
void ivec(void) { ifail(); }

void run(void) {
	UINT	fl, na;

	for (Ip = 0;;) {
		Op = Text[Ip++];
		if (Op & IARFLAG) {
			fl = Op;
			Op = Text[Ip++];
			na = (fl & IARMASK) >> IAMOFF;
			if (na) {
				switch ((fl & IL1MASK) >> IL1OFF) {
				case 0:	A1 = Text[Ip++];
					if (A1 > 0x7F) A1 = (-1 & ~0x7F) | A1;
					break;
				case 1:	A1 = (Text[Ip] << 8) | Text[Ip+1];
					Ip += 2;
					if (A1 > 0x7FFF) A1 = (-1 & ~0x7FFF)
						| A1;
					break;
				case 3:	A1 = (Text[Ip] << 24) |
						(Text[Ip+1] << 16) |
						(Text[Ip+2] << 8) | Text[Ip+3];
#ifndef HOST16
					if (A1 > 0x7FFFFFFFL)
						A1 = (-1L & ~0x7FFFFFFFL) | A1;
#endif
					Ip += 4;
					break;
				default:
					fail("bad L1 value", NULL);
					break;
				}
				if (na > 1) {
					switch ((fl & IL2MASK) >> IL2OFF) {
					case 0:	A2 = Text[Ip++];
						if (A2 > 0x7F)
							A2 = (-1 & ~0x7F) | A2;
						break;
					case 1:	A2 = (Text[Ip] << 8) |
							Text[Ip+1];
						Ip += 2;
						if (A2 > 0x7FFF)
							A2 = (-1 & ~0x7FFF)
								| A2;
						break;
					case 3:	A2 = (Text[Ip] << 24) |
							(Text[Ip+1] << 16) |
							(Text[Ip+2] << 8) |
							Text[Ip+3];
						Ip += 4;
#ifndef HOST16
						if (A2 > 0x7FFFFFFFL)
							A2 =
							(-1L & ~0x7FFFFFFFL)
							| A2;
#endif
						break;
					default:
						fail("bad L2 value", NULL);
						break;
					}
				}
			}
		}
		if (Op >= IENDOFSET) ifail();
		U1 = A1;
		(*Tcode_instructions[Op])();
	}
}

void version(void) {
	printf("TXX -- Virtual Tcode Machine, Version 6.3 (Tcode8)\n");
	printf("Copyright (C) 1997-2004 Nils M Holm.\n");
	printf("Cell size  = %d bits\n", BPW*8);
	printf("Extensions = T3X");
#ifdef EXT_SYSTEM
	printf(", System");
#endif
#ifdef EXT_TTYCTL
	printf(", TTYCtl");
#endif
	fputc('\n', stdout);
	exit(0);
}

int main(int argc, char **argv) {
	Argcount = argc-1;
	Argvec = argv+1;
	init();
	if (argc < 2) version();
	load(argv[1]);
	run();
	return 0;
}
