/*
 *	SYS -- Runtime Support Routines for T3X under Unix.
 *	Copyright (C) 1999,2000,2002,2003,2004 Nils M Holm.
 *	All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include "../compiler/txx.h"

#ifdef EXT_SYSTEM

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <time.h>

#ifdef COHERENT
 #include <poll.h>
 #define HAVE_POLL
 #ifndef S_IRUSR
  #define V7MODE
 #endif
 #define WEXITSTATUS(r)		(((r)>>8)&255)
 #define WIFEXITED(r)		(((r)&255)==0)
 #define DIRENT_NAMELEN(X)	(strlen((X)->d_name))
#else
 #include <sys/wait.h>
 #include <sys/time.h>
#endif

#ifdef __FreeBSD__
 #define FDSET_TYPE		struct fd_set
 #define DIRENT_NAMELEN(X)	((X)->d_namlen)
 #define HAVE_SELECT
#endif

#ifdef __NetBSD__
 #define FDSET_TYPE		struct fd_set
 #define DIRENT_NAMELEN(X)	((X)->d_namlen)
 #define HAVE_SELECT
#endif

#ifdef __linux__
 #define FDSET_TYPE		fd_set
 #define DIRENT_NAMELEN(X)	(strlen((X)->d_name))
 #define HAVE_SELECT
#endif

#ifdef V7MODE
 #define S_IRUSR	S_IREAD
 #define S_IWUSR	S_IWRITE
 #define S_IXUSR	S_IEXEC
 #define S_IRGRP	(S_IREAD>>3)
 #define S_IWGRP	(S_IWRITE>>3)
 #define S_IXGRP	(S_IEXEC>>3)
 #define S_IROTH	(S_IREAD>>6)
 #define S_IWOTH	(S_IWRITE>>6)
 #define S_IXOTH	(S_IEXEC>>6)
#endif

#define uchar	unsigned char
#define ushort	unsigned short
#define uint	unsigned int
#define ulong	unsigned long

#define NDIR	64
DIR	*Dirfd[NDIR];

#define MAXARGS	128

#define NPROC	64
int	Procd[NPROC];

#define RDOK	0x01
#define WROK	0x02
#define EXOK	0x04
#define ISDIR	0x08

XCELL system_init S0() {
	int	i;

	for (i=0; i<NDIR; i++) Dirfd[i] = NULL;
	for (i=0; i<NPROC; i++) Procd[i] = -1;
	return 0;
}

XCELL system_fini S0() {
	return 0;
}

int nextprocd(pid) {
	int	i;

	for (i=1; i<NPROC; i++) if (Procd[i] == -1) {
		Procd[i] = pid;
		return i;
	}
	write(2, "SYSTEM: out of process descriptors\n", 35);
	_exit(-1);
}

XCELL system_pipe S1(XCELL *fds) {
	int	a[2], r;

	r = pipe(a);
	fds[0] = a[0];
	fds[1] = a[1];
	return r;
}

XCELL system_dup S1(XCELL fd) {
	return dup(fd);
}

XCELL system_dup2 S2(XCELL ofd, XCELL nfd) {
	return dup2(ofd, nfd);
}

XCELL system_rdchk S1(XCELL fd) {
#ifdef HAVE_SELECT
	FDSET_TYPE	fdset;
	struct timeval	z;

	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);
	z.tv_sec = z.tv_usec = 0L;
	return (select(fd+1, &fdset, NULL, NULL, &z) == 1)? -1: 0;
#endif
#ifdef HAVE_POLL
	struct pollfd	p;

	p.fd = fd;
	p.events = POLLIN;
	return poll(&p, 1, 0);
#endif
	write(2, "SYSTEM.RDCHK: not implemented\n", 31);
	_exit(-1);
}

XCELL system_mkdir S1(char *s) {
	return mkdir(s, 0755);
}

XCELL system_chdir S1(char *s) {
	return chdir(s);
}

XCELL system_rmdir S1(char *s) {
	return rmdir(s);
}

XCELL system_getdir S2(char *buf, XCELL len) {
	if (len < 65) return -1;
	if (getcwd(buf, len) == NULL) return -1;
	return strlen(buf);
}

XCELL system_opendir S1(char *s) {
	int	i;

	for (i=0; i<NDIR; i++) if (Dirfd[i] == NULL) break;
	if (i >= NDIR) return -1;
	if ((Dirfd[i] = opendir(s)) == NULL) return -1;
	return i;
}

XCELL system_readdir S3(XCELL dd, char *b, XCELL n) {
	struct dirent	*dp;
	int		k;

	if (Dirfd[dd] == NULL) return -1;
	dp = readdir(Dirfd[dd]);
	if (dp == NULL) return -1;
	k = DIRENT_NAMELEN(dp);
	k = k>=n? n-1: k;
	memcpy(b, dp->d_name, k);
	b[k] = 0;
	return k;
}

XCELL system_closedir S1(XCELL dd) {
	DIR	*old;

	if (Dirfd[dd] == 0) return -1;
	old = Dirfd[dd];
	Dirfd[dd] = 0;
	return closedir(old);
}

uint accessmap(struct stat *sp) {
	int	uid = getuid(), gid = getgid();
	int	bits = 0;

	if (uid == 0) bits = RDOK|WROK;
	if (sp->st_uid == uid) {
		if (sp->st_mode & S_IRUSR) bits |= RDOK;
		if (sp->st_mode & S_IWUSR) bits |= WROK;
		if (sp->st_mode & S_IXUSR) bits |= EXOK;
	}
	if (sp->st_gid == gid) {
		if (sp->st_mode & S_IRGRP) bits |= RDOK;
		if (sp->st_mode & S_IWGRP) bits |= WROK;
		if (sp->st_mode & S_IXGRP) bits |= EXOK;
	}
	if (sp->st_mode & S_IROTH) bits |= RDOK;
	if (sp->st_mode & S_IWOTH) bits |= WROK;
	if (sp->st_mode & S_IXOTH) bits |= EXOK;
	return bits;
}

XCELL system__stat S2(char *name, uchar *b) {
	struct t3x_stat {
		short	dev;
		short	ino;
		short	mode;
		short	nlink;
		short	uid;
		short	gid;
		short	rdev;
		short	size;
		short	ext64;
		short	mt_year;
		short	mt_mon_day;
		short	mt_hr_min;
		short	mt_sec_hs;
	};
	struct stat	st;
	struct tm	*tp;

	if (stat(name, &st) < 0) return -1;
	b[0] = (uchar) (st.st_dev >> 8);
	b[1] = (uchar) st.st_dev;
	b[2] = (uchar) (st.st_ino >> 8);
	b[3] = (uchar) st.st_ino;
	b[4] = (uchar) 0;
	b[5] = (uchar) (st.st_mode & S_IFDIR? ISDIR: 0) | accessmap(&st);
	b[6] = (uchar) (st.st_nlink >> 8);
	b[7] = (uchar) st.st_nlink;
	b[8] = (uchar) (st.st_uid >> 8);
	b[9] = (uchar) st.st_uid;
	b[10] = (uchar) (st.st_gid >> 8);
	b[11] = (uchar) st.st_gid;
	b[12] = (uchar) (st.st_rdev >> 8);
	b[13] = (uchar) st.st_rdev;
	b[14] = (uchar) (st.st_size >> 8);
	b[15] = (uchar) st.st_size;
	b[16] = (uchar) (st.st_size / (64L*1024) >> 8);
	b[17] = (uchar) (st.st_size / (64L*1024));
	tp = localtime(&st.st_mtime);
	b[18] = (tp->tm_year+1900) / 100;
	b[19] = tp->tm_year % 100;
	b[20] = tp->tm_mon+1;
	b[21] = tp->tm_mday;
	b[22] = tp->tm_hour;
	b[23] = tp->tm_min;
	b[24] = tp->tm_sec;
	b[25] = 0;
	return 0;
}

int _system_wait(void) {
	int	r;

	wait(&r);
	if (WIFEXITED(r)) return WEXITSTATUS(r);
	return -1;
}

XCELL system_wait S1(XCELL pid) {
	int	r;

	if (Procd[pid] < 0) return -1;
	r = _system_wait();
	Procd[pid] = -1;
	return r;
}

XCELL system__spawn S3(char *p, char **av, XCELL w) {
	int	r, i;
	char	*av2[MAXARGS];

	for (i = 0; av[i]; i++)
		if (i >= MAXARGS-1) return -1;
	for (++i; i>0; i--) av2[i] = av[i-1];
	av2[0] = p;
	if ((r = fork()) < 0) return -1;
	if (r) {
		if (w) return _system_wait();
	}
	else {
		execv(p, av2);
		_exit(-1);
	}
	return nextprocd(r);
}

XCELL system_fork S0() {
	int	r;

	r = fork();
	if (!r || r < 0) return r;
	return nextprocd(r);
}

XCELL system_kill S2(XCELL pid, XCELL n) {
	int	r;

	if (Procd[pid] < 0) return -1;
	r = kill(Procd[pid], n);
	if (r < 0) return r;
	if (n) Procd[pid] = -1;
	return r;
}

void convtime(time_t t, char *b) {
	int	mlen[13] = { 0, 31,28,31,30,31,30,31,31,30,31,30,31 };
	int	yr, mon, day, hr, min, sec;
	long	dt;

	yr = 1970;
	while (1) {
		dt = (!(yr&3)? 366: 365) * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		yr++;
	}
	mon = 1;
	mlen[2] = !(yr&3)? 29: 28;
	while (mon<12) {
		dt = mlen[mon] * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		mon++;
	}
	day = (int) (t / 86400L + 1);
	t = (int) (t % 86400L);
	hr = (int) (t / 3600);
	t = (int) (t % 3600);
	min = (int) (t / 60);
	sec = (int) (t % 60);
	b[0] = yr / 100;
	b[1] = yr % 100;
	b[2] = mon;
	b[3] = day;
	b[4] = hr;
	b[5] = min;
	b[6] = sec;
	b[7] = 0;
}

XCELL system_time S1(char *b) {
	time_t	t;

	time(&t);
	convtime(t, b);
	return 0;
}

#ifndef LIBRARY

struct infproc	InfSystem[20] = {
	{ "system__spawn",	system__spawn	},
	{ "system__stat",	system__stat	},
	{ "system_chdir",	system_chdir	},
	{ "system_closedir",	system_closedir	},
	{ "system_dup",		system_dup	},
	{ "system_dup2",	system_dup2	},
	{ "system_fini",	system_fini	},
	{ "system_fork",	system_fork	},
	{ "system_getdir",	system_getdir	},
	{ "system_init",	system_init	},
	{ "system_kill",	system_kill	},
	{ "system_mkdir",	system_mkdir	},
	{ "system_opendir",	system_opendir	},
	{ "system_pipe",	system_pipe	},
	{ "system_rdchk",	system_rdchk	},
	{ "system_readdir",	system_readdir	},
	{ "system_rmdir",	system_rmdir	},
	{ "system_time",	system_time	},
	{ "system_wait",	system_wait	},
	{ NULL,			NULL		}
};

#endif

#else	/* EXT_SYSTEM */

#ifndef LIBRARY
struct infproc	InfSystem[MAXINFPROC] = { { 0L, 0L } };
#endif

#endif	/* EXT_SYSTEM */
