/*
 *	TTY -- Terminal I/O Routines for unixish Systems.
 *	Copyright (C) 1996-2004 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 *
 *	Using POSIX-style TERMIOS is preferred on systems which
 *	support it. On older systems, you may try V7 SGTTY calls
 *	by defining USE_SGTTY .
 */

#include "../compiler/txx.h"

#ifdef EXT_TTYCTL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef USE_SGTTY
 #define TTYSTATE	struct sgttyb
 #include <sgtty.h>
#else
 #define TTYSTATE	struct termios
 #include <termios.h>
#endif

#ifndef COHERENT
 #include <sys/time.h>
#endif

#ifdef COHERENT
 #include <poll.h>
 #define HAVE_POLL
 #define TPUTS3_TYPE	int(*)()
 char *tgetstr();
 #include <term.h>
#endif

#ifdef __FreeBSD__
 #define FDSET_TYPE	struct fd_set
 #define HAVE_TPARM
 #define HAVE_SELECT
 #define TPUTS3_TYPE	int(*)()
 #if __FreeBSD__ >= 4
  #include <curses.h>
  #include <term.h>
 #else
  #include <termcap.h>
 #endif
#endif

#ifdef __NetBSD__
 #define HAVE_SELECT
 #define FDSET_TYPE	struct fd_set
 #define TPUTS3_TYPE	int(*)()
 #include <termcap.h>
#endif

#ifdef __linux__
 #define HAVE_SELECT
 #define FDSET_TYPE	fd_set
 #define TPUTS3_TYPE	int(*)()
 #include <termcap.h>
#endif

#define K_HOME	327
#define K_WLFT	371
#define K_LEFT	331
#define K_RGHT	333
#define K_WRGT	372
#define K_END	335
#define K_BKSP	'\b'
#define K_DEL	339
#define K_KILL	127
#define K_INS	338
#define K_CR	'\r'
#define K_UP	328
#define K_DOWN	336
#define K_ESC	'\033'
#define K_PGUP	329
#define K_PGDN	337
#define K_CPGU	388
#define K_CPGD	374
#define K_F1	315
#define K_F2	316
#define K_F3	317
#define K_F4	318
#define K_F5	319
#define K_F6	320
#define K_F7	321
#define K_F8	322
#define K_F9	323
#define K_F10	324

#define CE	0
#define CL	1
#define CM	2
#define CS	3
#define SF	4
#define SR	5
#define SE	6
#define SO	7

static char	*tc[9][3] = {
	{	"ce",	NULL,	"erase line"			},
	{	"cl",	NULL,	"erase display"			},
	{	"cm",	NULL,	"cursor move"			},
	{	"cs",	NULL,	"change scrolling region"	},
	{	"sf",	NULL,	"scroll forward"		},
	{	"sr",	NULL,	"scroll reverse"		},
	{	"se",	NULL,	"end standout"			},
	{	"so",	NULL,	"standout"			},
	{	NULL	}
};

struct keycode {
	char	*name,
		*seq;
	int	code;
	char	*desc;
} kc[] = {
	{	"**",	"\033[B",	K_DOWN,	"vt100 down"	},
	{	"**",	"\033[D",	K_LEFT,	"vt100 left"	},
	{	"**",	"\033[C",	K_RGHT,	"vt100 right"	},
	{	"**",	"\033[A",	K_UP,	"vt100 up"	},
	{	"kd",	"\033[B",	K_DOWN,	"key down"	},
	{	"kl",	"\033[D",	K_LEFT,	"key left"	},
	{	"kr",	"\033[C",	K_RGHT,	"key right"	},
	{	"ku",	"\033[A",	K_UP,	"key up"	},
	{	"k1",	"\033Oq",	K_F1,	"key F1"	},
	{	"k2",	"\033Or",	K_F2,	"key F2"	},
	{	"k3",	"\033Os",	K_F3,	"key F3"	},
	{	"k4",	"\033Ot",	K_F4,	"key F4"	},
	{	"k5",	"\033Ou",	K_F5,	"key F5"	},
	{	"k6",	"\033Ov",	K_F6,	"key F6"	},
	{	"k7",	"\033Ow",	K_F7,	"key F7"	},
	{	"k8",	"\033Ox",	K_F8,	"key F8"	},
	{	"k9",	"\033Oy",	K_F9,	"key F9"	},
	{	"k0",	"\033Op",	K_F10,	"key F10"	},
	{	"kP",	"\033Or",	K_PGUP,	"key prev"	},
	{	"kN",	"\033Oq",	K_PGDN,	"key next"	},
	{	"kh",	"\033[H",	K_HOME,	"key home"	},
	{	"kH",	"\033[F",	K_END,	"key eol"	},
	{	"kI",	"\033On",	K_INS,	"key insert"	},
	{	"kD",	"\033D",	K_DEL,	"key delete"	},
	{	NULL,	NULL,		0,	NULL		}
};

#define MAXLEVEL	64

static TTYSTATE		rawbits, oldbitsv[MAXLEVEL];
static int		level = 0;

#define KBUFLEN	64
static char	key_buf[KBUFLEN], key_ptr = 0, key_len = 0;

int	tty_infd,
	tty_outfd,
	tty_cols,
	tty_lines,
	tty_attr,
	tty_x,
	tty_y;
int	tty_soflag = 0;

int	Ansicolor;

unsigned char	**Chrmap, **Atrmap;

#ifndef HAVE_TPARM

/* quick and dirty tparm() implementation
 * for systems that don't have it
 */
static char	tparm_buf[81];
char *tparm(char *s, int a1, int a2) {
	int	i, j, ax;

	ax = 0;
	for (i=j=0; s[i]; i++) {
		if (j > 79) return "OOPS";
		if (s[i] == '%') {
			switch (s[++i]) {
			case 'i':
				a1++; a2++;
				break;
			case 'd':
				sprintf(&tparm_buf[j], "%d", ax? a2: a1);
				j += strlen(&tparm_buf[j]);
				ax = 1;
				break;
			default:
				tparm_buf[j++] = s[i];
			}
		}
		else {
			tparm_buf[j++] = s[i];
		}
	}
	tparm_buf[j] = 0;
	return tparm_buf;
}

#endif /* HAVE_TPARM */

static int _ttyctl_writec(c)
int	c;
{
	char	ch = c;

	write(tty_outfd, &ch, 1);
	return c;
}

XCELL ttyctl_writec S1(XCELL c) {
	char	ch = c;

	Chrmap[tty_y][tty_x] = c;
	Atrmap[tty_y][tty_x] = tty_attr;
	tty_x++;
	write(tty_outfd, &ch, 1);
	return c;
}

XCELL ttyctl_writes S1(char *s) {
	int	k = strlen(s);

	memcpy(&Chrmap[tty_y][tty_x], s, k);
	memset(&Atrmap[tty_y][tty_x], tty_attr, k);
	tty_x += k;
	write(tty_outfd, s, strlen(s));
	return 0;
}

XCELL ttyctl_clear S0() {
	int	i;

	for (i=0; i<tty_lines; i++) {
		memset(Chrmap[i], ' ', tty_cols);
		memset(Atrmap[i], tty_attr, tty_cols);
	}
	tty_x = 0;
	tty_y = 0;
	tputs(tc[CL][1], 1, (TPUTS3_TYPE)_ttyctl_writec);
	return 0;
}

XCELL ttyctl_clreol S0() {
	memset(&Chrmap[tty_y][tty_x], ' ', tty_cols-tty_x);
	memset(&Atrmap[tty_y][tty_x], tty_attr, tty_cols-tty_x);
	tputs(tc[CE][1], 1, (TPUTS3_TYPE)_ttyctl_writec);
	return 0;
}

int _ttyctl_color(int c) {
	int	ctab[8] = { '0', '4', '2', '6', '1', '5', '3', '7' };
	int	f, b, h;
	char	buf[40];

	if (tty_attr == c) return;
	tty_attr = c;
	f = ctab[c & 7];
	b = ctab[(c>>4) & 7];
	if (Ansicolor) {
		h = ((c&8) == 8) + '0';
		sprintf(buf, "\033[%c;3%c;4%cm", (char) h, (char) f, (char) b);
		tputs(buf, 1, (TPUTS3_TYPE)_ttyctl_writec);
	}
	else {
		/*
		 * Coherent hack: SO (^[[7m) selects reverse video
		 * relative to current attribute. Hence emitting
		 * SO twice selects normal video.
		 */
		if (tty_soflag != (b>f)) {
			if (b > f)
				tputs(tc[SO][1], 1,
					(TPUTS3_TYPE)_ttyctl_writec);
			else
				tputs(tc[SE][1], 1,
					(TPUTS3_TYPE)_ttyctl_writec);
			tty_soflag = (b>f);
		}
	}
}

XCELL ttyctl_color S1(XCELL c) {
	_ttyctl_color(c);
	return 0;
}

int _ttyctl_query(ms) {
#ifdef HAVE_SELECT
	FDSET_TYPE	fdset;
	struct timeval	z;

	FD_ZERO(&fdset);
	FD_SET(tty_infd, &fdset);
	z.tv_sec = 0L;
	z.tv_usec = ms * 1000L;
	return select(tty_infd+1, &fdset, NULL, NULL, &z) == 1;
#endif
#ifdef HAVE_POLL
	struct pollfd	p;

	p.fd = tty_infd;
	p.events = POLLIN;
	return poll(&p, 1, ms);
#endif
	write(2, "TTYCTL.QUERY: not implemented\n", 30);
	_exit(-1);
}

XCELL ttyctl_query S0() {
	return _ttyctl_query(0);
}

XCELL ttyctl_readc S0() {
	char		ch;
	int		i;

rescan:
	if (key_len > key_ptr) {
		return key_buf[key_ptr++];
	}
	read(tty_infd, &ch, 1);
	if (ch != K_ESC) return ch;
	if (!_ttyctl_query(300)) return ch;
	key_buf[0] = K_ESC; key_ptr = 1;
	while (1) {
		if (!_ttyctl_query(300)) break;
		if (key_ptr >= KBUFLEN-1) break;
		read(tty_infd, &ch, 1);
		key_buf[key_ptr++] = ch;
		key_buf[key_ptr] = 0;
		for (i=0; kc[i].name; i++) {
			if (kc[i].seq && !strcmp(key_buf, kc[i].seq)) {
				key_len = key_ptr = 0;
				return kc[i].code;
			}
		}
	}
	key_len = key_ptr; key_ptr = 0;
	goto rescan;
}

int _ttyctl_move(int x, int y) {
	tty_x = x;
	tty_y = y;
	tputs(tgoto(tc[CM][1], tty_x, tty_y), 1, (TPUTS3_TYPE)_ttyctl_writec);
}

XCELL ttyctl_move S2(XCELL x, XCELL y) {
	_ttyctl_move(x, y);
	return 0;
}

static void redraw(int top, int bottom) {
	int	i, j, k;

	for (i=top; i<=bottom; i++) {
		j = 0;
		k = 0;
		_ttyctl_move(0, i);
		while (k < tty_cols) {
			if (tty_attr != Atrmap[i][j])
				_ttyctl_color(Atrmap[i][j]);
			while (j < tty_cols && tty_attr == Atrmap[i][j]) j++;
			write(tty_outfd, &Chrmap[i][k], j-k);
			k = j;
		}
	}
}

XCELL ttyctl_scroll S2(XCELL top, XCELL bot) {
	int	i;

	if (top == 0 && bot == tty_lines-1) {
		_ttyctl_move(0, tty_lines-1);
		tputs(tc[SF][1], bot-top+1, (TPUTS3_TYPE)_ttyctl_writec);
	}
	else if (tc[CS][1]) {
		tputs(tparm(tc[CS][1], top, bot), bot-top+1,
			(TPUTS3_TYPE)_ttyctl_writec);
		_ttyctl_move(0, bot);
		tputs(tc[SF][1], bot-top+1, (TPUTS3_TYPE)_ttyctl_writec);
		tputs(tparm(tc[CS][1], 0, tty_lines-1), tty_lines,
			(TPUTS3_TYPE)_ttyctl_writec);
	}
	else {
		for (i=top; i<bot; i++) {
			memcpy(Chrmap[i], Chrmap[i+1], tty_cols);
			memcpy(Atrmap[i], Atrmap[i+1], tty_cols);
		}
		memset(Chrmap[bot], ' ', tty_cols);
		memset(Atrmap[bot], tty_attr, tty_cols);
		redraw(top, bot);
	}
	return 0;
}

XCELL ttyctl_rscroll S2(XCELL top, XCELL bot) {
	int	i;

	if (top == 0 && bot == tty_lines-1) {
		_ttyctl_move(0, 0);
		tputs(tc[SR][1], bot-top+1, (TPUTS3_TYPE)_ttyctl_writec);
	}
	else if (tc[CS][1]) {
		tputs(tparm(tc[CS][1], top, bot), bot-top+1,
			(TPUTS3_TYPE)_ttyctl_writec);
		_ttyctl_move(0, top);
		tputs(tc[SR][1], bot-top+1, (TPUTS3_TYPE)_ttyctl_writec);
		tputs(tparm(tc[CS][1], 0, tty_lines-1), tty_lines,
			(TPUTS3_TYPE)_ttyctl_writec);
	}
	else {
		for (i=bot; i>top; i--) {
			memcpy(Chrmap[i], Chrmap[i-1], tty_cols);
			memcpy(Atrmap[i], Atrmap[i-1], tty_cols);
		}
		memset(Chrmap[top], ' ', tty_cols);
		memset(Atrmap[top], tty_attr, tty_cols);
		redraw(top, bot);
	}
	return 0;
}

unsigned char **allocmap(int co, int li) {
	unsigned char	**p;
	char		*msg;
	int		i;

	msg = "TTYCTL: not enough memory for screen map\n";
	if ((p = (unsigned char **)
		malloc(li*sizeof(unsigned char *))) == NULL)
	{
		fprintf(stderr, msg);
		_exit(-1);
	}
	for (i=0; i<li; i++) {
		if ((p[i] = (unsigned char *) malloc(co)) == NULL) {
			fprintf(stderr, msg);
			_exit(-1);
		}
	}
	return p;
}

void freemap(unsigned char **map) {
	int	i;

	for (i=0; i<tty_lines; i++) free(map[i]);
	free(map);
}

XCELL ttyctl_init S0() {
	char		*tname, *tp, tbuf[1024], *t;
	static char	tsbuf[1024];
	int		i;

	if (level >= MAXLEVEL-1) {
		fprintf(stderr, "TTYCTL: too many nested init()s\n");
		exit(EXIT_FAILURE);  
	}
	if (!level) {
		if (	(tty_infd = open("/dev/tty", 0)) < 0 ||
			(tty_outfd = open("/dev/tty", 1)) < 0
		) {
			fprintf(stderr,
				"TTYCTL: failed to connect to /dev/tty\n");
			exit(EXIT_FAILURE);
		}
	}
	if ((tname = getenv("TERM")) == NULL) {
		fprintf(stderr, "TTYCTL: the TERM variable is not set\n");
		exit(EXIT_FAILURE);
	}
	t = getenv("ANSICOLOR");
	if (t != NULL) {
		Ansicolor = !strcmp(t, "no")? 0:
			!strcmp(t, "No")? 0:
			!strcmp(t, "NO")? 0: 1;
	}
	else {
		Ansicolor = strstr(tname, "color") != NULL ||
			memcmp(tname, "xterm", 5) == 0;
	}
	switch(tgetent(tbuf, tname)) {
		case -1: fprintf(stderr,
			 "TTYCTL: could not open termcap database");
			 exit(EXIT_FAILURE);
		case 0:  fprintf(stderr,
			"TTYCTL: unknown terminal type: %s\n", tname);
			 exit(EXIT_FAILURE);
	}
	if (	(tty_cols = tgetnum("co")) == -1 ||
		(tty_lines = tgetnum("li")) == -1
	) {
		fprintf(stderr, "TTYCTL: cannot determine screen size\n");
		exit(EXIT_FAILURE);
	}
	Chrmap = allocmap(tty_cols, tty_lines);
	Atrmap = allocmap(tty_cols, tty_lines);
	tp = tsbuf;
	for (i=0; tc[i][0]; i++) {
		if (	(tc[i][1] = tgetstr(tc[i][0], &tp)) == NULL &&
			strcmp(tc[i][0], "cs")
		) {
			fprintf(stderr,
			"TTYCTL: your terminal is lacking the"
			" '%s' capability\n", tc[i][2]);
			exit(EXIT_FAILURE);
		}
	}
	for (i=0; kc[i].name; i++) {
		if ((t = tgetstr(kc[i].name, &tp)) != NULL)
			kc[i].seq = t;
	}
	tty_x = 0;
	tty_y = 0;
	tty_attr = 0x07;
	if (level == 0) {
		for (i=0; i<tty_lines; i++) {
			memset(Chrmap[i], ' ', tty_cols);
			memset(Atrmap[i], tty_attr, tty_cols);
		}
	}
	key_len = key_ptr = 0;
#ifdef USE_SGTTY
	gtty(tty_infd, &oldbitsv[level]);
	gtty(tty_infd, &rawbits);
	rawbits.sg_flags |= RAW;
	rawbits.sg_flags &= ~(ECHO|CRMOD);
#else
	tcgetattr(tty_infd, &oldbitsv[level]);
	tcgetattr(tty_infd, &rawbits);
	cfmakeraw(&rawbits);
#endif
	level++;
	return 0;
}

void _ttyctl_mode(int rawmode) {
#ifdef USE_SGTTY
	stty(tty_infd, rawmode? &rawbits: &oldbitsv[level-1]);
#else
	tcsetattr(tty_infd, TCSANOW|TCSADRAIN, rawmode? &rawbits:
		&oldbitsv[level-1]);
#endif
}

XCELL ttyctl_mode S1(XCELL rawmode) {
	_ttyctl_mode(rawmode);
	return 0;
}

XCELL ttyctl_fini S0() {
	if (level < 0) {
		fprintf(stderr, "TTYCTL: redundant fini()\n");
		exit(EXIT_FAILURE);
	}
	_ttyctl_mode(0);
	level--;
	if (!level) {
		freemap(Chrmap);
		freemap(Atrmap);
		close(tty_infd);
		close(tty_outfd);
	}
	return 0;
}

XCELL ttyctl_lines S0() { return tty_lines; }

XCELL ttyctl_columns S0() { return tty_cols; }

XCELL ttyctl_colors S0() { return Ansicolor; }

#ifndef LIBRARY

struct infproc	InfTTYCtl[17] = {
	{ "ttyctl_clear",	ttyctl_clear	},
	{ "ttyctl_clreol",	ttyctl_clreol	},
	{ "ttyctl_color",	ttyctl_color	},
	{ "ttyctl_colors",	ttyctl_colors	},
	{ "ttyctl_columns",	ttyctl_columns	},
	{ "ttyctl_fini",	ttyctl_fini	},
	{ "ttyctl_init",	ttyctl_init	},
	{ "ttyctl_lines",	ttyctl_lines	},
	{ "ttyctl_mode",	ttyctl_mode	},
	{ "ttyctl_move",	ttyctl_move	},
	{ "ttyctl_query",	ttyctl_query	},
	{ "ttyctl_readc",	ttyctl_readc	},
	{ "ttyctl_rscroll",	ttyctl_rscroll	},
	{ "ttyctl_scroll",	ttyctl_scroll	},
	{ "ttyctl_writec",	ttyctl_writec	},
	{ "ttyctl_writes",	ttyctl_writes	},
	{ NULL,			NULL		}
};

#endif

#else	/* EXT_TTYCTL */

#ifndef LIBRARY
struct infproc	InfTTYCtl[MAXINFPROC] = { { 0L, 0L } };
#endif

#endif	/* EXT_TTYCTL */
