/*
 *	TTY -- VT220 Routines for Plan 9.
 *	Copyright (C) 2004 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include "../compiler/txx.h"

#ifdef EXT_TTYCTL

#include <u.h>
#include <libc.h>

#define KBUFLEN	64
static char	key_buf[KBUFLEN], key_ptr = 0, key_len = 0;

int	tty_infd,
	tty_outfd,
	tty_cols,
	tty_lines,
	tty_attr,
	tty_x,
	tty_y;
int	tty_soflag = 0;

#define K_HOME	327
#define K_WLFT	371
#define K_LEFT	331
#define K_RGHT	333
#define K_WRGT	372
#define K_END	335
#define K_BKSP	'\b'
#define K_DEL	339
#define K_KILL	127
#define K_INS	338
#define K_CR	'\r'
#define K_UP	328
#define K_DOWN	336
#define K_ESC	'\033'
#define K_PGUP	329
#define K_PGDN	337
#define K_CPGU	388
#define K_CPGD	374
#define K_F1	315
#define K_F2	316
#define K_F3	317
#define K_F4	318
#define K_F5	319
#define K_F6	320
#define K_F7	321
#define K_F8	322
#define K_F9	323
#define K_F10	324

struct keycode {
	char	*name,
		*seq;
	int	code;
	char	*desc;
} kc[] = {
	{	"**",	"\033[B",	K_DOWN,	"vt100 down"	},
	{	"**",	"\033[D",	K_LEFT,	"vt100 left"	},
	{	"**",	"\033[C",	K_RGHT,	"vt100 right"	},
	{	"**",	"\033[A",	K_UP,	"vt100 up"	},
	{	"kd",	"\033[B",	K_DOWN,	"key down"	},
	{	"kl",	"\033[D",	K_LEFT,	"key left"	},
	{	"kr",	"\033[C",	K_RGHT,	"key right"	},
	{	"ku",	"\033[A",	K_UP,	"key up"	},
	{	"k1",	"\033Oq",	K_F1,	"key F1"	},
	{	"k2",	"\033Or",	K_F2,	"key F2"	},
	{	"k3",	"\033Os",	K_F3,	"key F3"	},
	{	"k4",	"\033Ot",	K_F4,	"key F4"	},
	{	"k5",	"\033Ou",	K_F5,	"key F5"	},
	{	"k6",	"\033Ov",	K_F6,	"key F6"	},
	{	"k7",	"\033Ow",	K_F7,	"key F7"	},
	{	"k8",	"\033Ox",	K_F8,	"key F8"	},
	{	"k9",	"\033Oy",	K_F9,	"key F9"	},
	{	"k0",	"\033Op",	K_F10,	"key F10"	},
	{	"kP",	"\033Or",	K_PGUP,	"key prev"	},
	{	"kN",	"\033Oq",	K_PGDN,	"key next"	},
	{	"kh",	"\033[H",	K_HOME,	"key home"	},
	{	"kH",	"\033[F",	K_END,	"key eol"	},
	{	"kI",	"\033On",	K_INS,	"key insert"	},
	{	"kD",	"\033D",	K_DEL,	"key delete"	},
	{	nil,	nil,		0,	nil		}
};

static void tputs(char *s) {
	write(tty_outfd, s, strlen(s));
}

XCELL ttyctl_writec S1(XCELL c) {
	char	ch = c;

	tty_x++;
	write(tty_outfd, &ch, 1);
	return c;
}

XCELL ttyctl_writes S1(char *s) {
	int	k = strlen(s);

	tty_x += k;
	if (k) write(tty_outfd, s, k);
	return 0;
}

XCELL ttyctl_clear S0() {
	tty_x = 0;
	tty_y = 0;
	tputs("\033[H\033[2J");
	return 0;
}

XCELL ttyctl_clreol S0() {
	tputs("\033[K");
	return 0;
}

XCELL ttyctl_color S1(XCELL c) {
	int	ctab[8] = { '0', '4', '2', '6', '1', '5', '3', '7' };
	int	f, b;

	if (tty_attr == c) return 0;
	tty_attr = c;
	f = ctab[c & 7];
	b = ctab[(c>>4) & 7];
	if (tty_soflag != (b>f)) {
		if (b > f)
			tputs("\033[7m");
		else
			tputs("\033[m");
		tty_soflag = (b>f);
	}
	return 0;
}

XCELL ttyctl_query S0() {
	write(2, "TTYCTL.QUERY: not implemented\n", 30);
	exits("TTYCTL.QUERY: not implemented");
	return 0;
}

XCELL ttyctl_readc S0() {
	char		ch;
	int		i, match;

rescan:
	if (key_len > key_ptr) {
		return key_buf[key_ptr++];
	}
	read(tty_infd, &ch, 1);
	if (ch != '\033') return ch;
	key_buf[0] = '\033'; key_ptr = 1;
	while (1) {
		if (key_ptr >= KBUFLEN-1) break;
		read(tty_infd, &ch, 1);
		key_buf[key_ptr++] = ch;
		key_buf[key_ptr] = 0;
		match = 0;
		for (i=0; kc[i].name; i++) {
			if (kc[i].seq && !strcmp(key_buf, kc[i].seq)) {
				key_len = key_ptr = 0;
				return kc[i].code;
			}
			if (kc[i].seq && !strncmp(key_buf, kc[i].seq, key_ptr))
				match = 1;
		}
		if (!match) break;
	}
	key_len = key_ptr; key_ptr = 0;
	goto rescan;
}

void move(int x, int y) {
	char	b[100];

	tty_x = x;
	tty_y = y;
	sprint(b, "\033[%d;%dH", y+1, x+1);
	tputs(b);
}

XCELL ttyctl_move S2(XCELL x, XCELL y) {
	move(x, y);
	return 0;
}

XCELL ttyctl_scroll S2(XCELL top, XCELL bot) {
	char	b[100];

	if (top == 0 && bot == tty_lines-1) {
		move(0, tty_lines-1);
		tputs("\033D");
	}
	else {
		sprint(b, "\033[%d;%dr", top+1, bot+1);
		tputs(b);
		move(0, bot);
		tputs("\033D");
		tputs("\033[r");
	}
	return 0;
}

XCELL ttyctl_rscroll S2(XCELL top, XCELL bot) {
	char	b[100];

	if (top == 0 && bot == tty_lines-1) {
		move(0, 0);
		tputs("\033M");
	}
	else {
		sprint(b, "\033[%d;%dr", top+1, bot+1);
		tputs(b);
		move(0, top);
		tputs("\033M");
		tputs("\033[r");
	}
	return 0;
}

XCELL ttyctl_init S0() {
	tty_infd = 0;
	tty_outfd = 1;
	tty_x = 0;
	tty_y = 0;
	tty_lines = 24;
	tty_cols = 80;
	tty_attr = 0x07;
	tty_soflag = 0;
	key_len = key_ptr = 0;
	tputs("\033[H\033[2J");
	return 0;
}

XCELL ttyctl_mode S1(XCELL *) {
	return 0;
}

XCELL ttyctl_fini S0() {
	return 0;
}

XCELL ttyctl_lines S0() { return tty_lines; }

XCELL ttyctl_columns S0() { return tty_cols; }

XCELL ttyctl_colors S0() { return 0; }

#ifndef LIBRARY

struct infproc	InfTTYCtl[17] = {
	{ "ttyctl_clear",	ttyctl_clear	},
	{ "ttyctl_clreol",	ttyctl_clreol	},
	{ "ttyctl_color",	ttyctl_color	},
	{ "ttyctl_colors",	ttyctl_colors	},
	{ "ttyctl_columns",	ttyctl_columns	},
	{ "ttyctl_fini",	ttyctl_fini	},
	{ "ttyctl_init",	ttyctl_init	},
	{ "ttyctl_lines",	ttyctl_lines	},
	{ "ttyctl_mode",	ttyctl_mode	},
	{ "ttyctl_move",	ttyctl_move	},
	{ "ttyctl_query",	ttyctl_query	},
	{ "ttyctl_readc",	ttyctl_readc	},
	{ "ttyctl_rscroll",	ttyctl_rscroll	},
	{ "ttyctl_scroll",	ttyctl_scroll	},
	{ "ttyctl_writec",	ttyctl_writec	},
	{ "ttyctl_writes",	ttyctl_writes	},
	{ nil,			nil		}
};

#endif

#else	/* EXT_TTYCTL */

#ifndef LIBRARY
struct infproc	InfTTYCtl[MAXINFPROC] = { { 0L, 0L } };
#endif

#endif	/* EXT_TTYCTL */
